'''
/Topic/lda.py
Implementation of LDA topic-model
'''
import re
import os
import pathlib
import gensim
import numpy as np
from womg_core.topic.lda import LDA
from womg_core.utils.utility_functions import count_files, read_docs
from womg_core.utils.distributions import random_powerlaw_vec

class LDAExtended(LDA):
    '''
    Class implementing Latent Dirichlet Allocation as topic model
    for topic distribution involved in tlt class model

    Attributes
    ----------
    numb_topics : int
        hidden
    items_keyw : dict
        dict of items in bow format

    Methods
    -------
    set_lda_mode : concrete
        sets the lda mode: reading or generating mode
    '''

    def __init__(self, numb_topics,
                 numb_docs,
                 docs_path,
                 items_descr,
                 progress_bar):
        super().__init__(numb_topics,
                         numb_docs,
                         items_descr,
                         progress_bar)
        self._docs_path = docs_path




    def fit(self):
        '''
        Pipeline for fitting the lda model

        1. define the lda mode : generative mode / reading mode
        2. train lda
        3. get the items descriptions (topic distribution for each item)
        4. get the items keywords (bow list for each item)
        '''
        mode = self.set_lda_mode()
        if mode == 'load':
            super().fit()

        if mode == 'get':
            lda_model = self.train_lda()
            self.topics_descript = self.get_topics_descript(lda_model)
            self.get_items_descript(path=self._docs_path, model=lda_model)
            self.get_items_keyw(path=self._docs_path)
        if mode == 'gen':
            lda_model = self.train_lda()
            self.topics_descript = self.get_topics_descript(lda_model)
            super().gen_items_descript()
            self.gen_items_keyw(model=lda_model)



    def set_lda_mode(self):
        '''
        Sets how lda has to work:
        reading a document folder or generating documents


        Parameters
        ----------
        path : string
            position of the document folder

        Returns
        -------
        reading : bool
            if True: it will read docs inside the given folder path or input folder
            if False: it will use lda for generating docs
        '''
        if self.numb_docs is None and self._docs_path is not None and self._items_descr is None:
            mode = 'get'
            path = pathlib.Path(self._docs_path)
            numb_docs = count_files(path)
            if numb_docs != 0:
                self.numb_docs = numb_docs
            else:
                print('No txt file in: ', path)
        elif self.numb_docs is not None and self._docs_path is not None and self._items_descr is None:
            mode = 'get'
        elif self.numb_docs is not None and self._docs_path is None and self._items_descr is None:
            mode = 'gen'
        else:
            mode = super().set_lda_mode()

        return mode


    def get_items_descript(self, path, model):
        '''
        Gets the topic distribution and puts it as attribute of the superclass
        reading documents from the docs folder path given as parameter

        Parameters
        ----------
        path : str
            path of the of the docs folder
        model : Gensim obj
            trained Gensim lda model
        '''
        print('Extracting topic distribution from docs in ', path)
        docs = read_docs(path, verbose=False)
        corpus = self.preprocess_texts(docs)
        gammas = {}
        item = 0
        for item in range(self.numb_docs):
            item_descript = model.get_document_topics(corpus[item], minimum_probability=0.)
            gammas[item] = np.array([i[1] for i in item_descript])
            item += 1
        if self.numb_docs == len(gammas.keys()):
            pass
        self.items_descript = gammas



    def get_items_keyw(self, path):
        '''
        Get the items keyword in a bow format
        '''
        docs = read_docs(path)
        preproc_docs = [gensim.parsing.preprocessing.remove_stopwords((sent[0].lower())) for sent in docs if len(sent) != 0]
        data_words = self.sent_to_words(preproc_docs, verbose=False)
        for item in range(self.numb_docs):
            self.items_keyw[item] = self.to_bow(data_words[item])


    def gen_items_keyw(self, model):
        '''
        Generates the items keyword in a bow format
        '''
        items_descript = self.items_descript
        topics_descript = self.get_topics_descript(model, mtrx_form=True)

        for item in range(self.numb_docs):
            item_keyw = []
            for _ in range(self.numb_words):
                multi_items = np.random.multinomial(1, items_descript[item], size=1)
                topic_index = np.where(multi_items == 1)[1]
                multi_topics = np.random.multinomial(1, topics_descript[topic_index][0], size=1)
                word_index = np.where(multi_topics == 1)[1]
                item_keyw.append(self.dictionary[word_index[0]])
            self.items_keyw[item] = self.to_bow(item_keyw)

    @staticmethod
    def get_topics_descript(model, mtrx_form=False):
        '''
        Getting the word distribution for each topic

        Parameters
        ----------
        model : gensim obj
            input model from gensim library e.g. lda model trained
        all : bool
            if True the complete word distribution for each topic is given
            in a (topics)x(words) matrix format
            if False a combination of most important words is given for each topic
        '''
        if mtrx_form:
            return model.get_topics()
        return model.print_topics()


    def to_bow(self, text):
        '''
        Returns the give doc in a bow format
        '''
        bow = self.dictionary.doc2bow(text)
        bow_dict = {}
        for word_index, freq in bow:
            bow_dict[self.dictionary[word_index]] = freq
        return bow_dict

    def preprocess_texts(self, docs):
        '''
        Preprocessing input texts: divides docs into words, bow format
        '''
        data = [gensim.parsing.preprocessing.remove_stopwords((sent[0].lower())) for sent in docs if len(sent) != 0]

        # Remove new line characters
        data = [re.sub(r'\s+', ' ', str(sent)) for sent in data]

        # Remove distracting single quotes
        data = [re.sub(r"\'", "", str(sent)) for sent in data]

        # Remove all the special characters
        data = [re.sub(r'\W', ' ', str(sent)) for sent in data]

        # remove all single characters
        data = [re.sub(r'\s+[a-zA-Z]\s+', ' ', str(sent)) for sent in data]

        # Remove single characters from the start
        data = [re.sub(r'\^[a-zA-Z]\s+', ' ', str(sent)) for sent in data]

        # Substituting multiple spaces with single space
        data = [re.sub(r'\s+', ' ', str(sent), flags=re.I) for sent in data]

        # Removing prefixed 'b'
        data = [re.sub(r'^b\s+', '', str(sent)) for sent in data]

        # Remove article
        data = [re.sub(r'the', '', str(sent)) for sent in data]

        # Remove to
        data = [re.sub(r'to', '', str(sent)) for sent in data]

        data_words = self.sent_to_words(data, verbose=False)
        corpus = [self.dictionary.doc2bow(word) for word in data_words]
        return corpus


    def sent_to_words(self, docs, verbose=True):
        '''
        Returns list of docs divided in words

        Parameters
        ----------
        docs : list
            list of lists: each entry is the list containing the document as a string

        Returns
        -------
        data_words : list
            list of lists: each entry is the list of words of one doc
        '''
        data_words = []
        if verbose:
            for sentence in self._progress_bar(docs):
                data_words.append(gensim.utils.simple_preprocess(str(sentence), deacc=True))
        else:
            for sentence in docs:
                data_words.append(gensim.utils.simple_preprocess(str(sentence), deacc=True))
        return list(data_words)


    def train_lda(self):
        '''
        Pre-train lda model on a saved corpus for infering the prior weights of
        the distributions given a number of topics, which correpsonds to the
        weights' dimension

        Returns gensim object of lda model

        Parameters
        ----------
        path : str
            path of the training corpus

        Returns
        -------
        gensim lda model object
        '''
        print('Training LDA model..')
        docs = read_docs(self._training_path)
        data_words = self.sent_to_words(docs, verbose=False)

        self.dictionary = gensim.corpora.Dictionary(data_words)
        corpus = self.preprocess_texts(docs)

        hash_name = 'trained_lda'+str(hash(str(self._training_path)+str(self.numb_topics)))+'.model'
        saved_model_fname = self.main_data_path/'topic_model'/hash_name
        if os.path.exists(os.path.abspath(saved_model_fname)):
            lda_model = gensim.models.ldamodel.LdaModel.load(os.path.abspath(saved_model_fname))
        else:
            lda_model = gensim.models.ldamodel.LdaModel(corpus=corpus,
                                                        id2word=self.dictionary,
                                                        num_topics=self.numb_topics,
                                                        random_state=100,
                                                        update_every=1,
                                                        chunksize=100,
                                                        passes=10,
                                                        alpha='auto',
                                                        per_word_topics=True)

            lda_model.save(os.path.abspath(saved_model_fname))

        return lda_model



'''
Reference implementation of node2vec.

Author: Aditya Grover

For more details, refer to the paper:
node2vec: Scalable Feature Learning for Networks
Aditya Grover and Jure Leskovec
Knowledge Discovery and Data Mining (KDD), 2016
'''
import sys
import click
import pathlib
import random
import numpy as np
import networkx as nx
from tqdm import tqdm
import tensorflow as tf
from womg.n2i.node2vec import node2vec, read_graph, save_emb

curr_path = str(pathlib.Path.cwd())

def n2i_nx_graph(nx_graph,
         weighted=False, directed=False,
         fast=False, seed=None,
         dimensions=128, walk_length=80,
         num_walks=10, window_size=10,
         iiter=1, workers=8,
         p=1, q=1,
         verbose=False,
         beta=0,
         prior='half_norm',
         alpha_value=2.,
         beta_value=2.):

    # seed
    if seed != None:
        random.seed(seed)
        np.random.seed(seed)
        tf.set_random_seed(seed)


    numb_nodes = nx.number_of_nodes(nx_graph)
    embeddings = node2vec(
                    weighted=weighted,
                    graph=nx_graph,
                    directed=directed,
                    p=p, q=q,
                    num_walks=num_walks,
                    walk_length=walk_length,
                    dimensions=dimensions,
                    window_size=window_size,
                    workers=workers,
                    iiter=iiter,
                    verbose=verbose,
                    beta=beta,
                    prior=prior,
                    alpha_value=alpha_value,
                    beta_value=beta_value)

    return embeddings

def n2i_main(topics=15,
         graph=None, fast=False,
         weighted=False, directed=False,
         output=curr_path, seed=None,
         dimensions=128, walk_length=80,
         num_walks=10, window_size=10,
         iiter=1, workers=8,
         p=1, q=1,
         beta=0,
         prior='half_norm',
         alpha_value=2.,
         beta_value=2.,
         verbose=False):
    '''



    Creates interests vector for each node using node2vec algorithm and
    saves interests in a file.

    3 steps:
        1. finding node2vec embeddings
        2. translation to positive axes of the embeddings
        3. reduction with NMF to _topics dimensions

    Parameters
    ----------
    - topics : int
        Number of topics in the topic model. Default 15. K<d
    - graph : str
        Input path of the graph edgelist
    - fast : bool
        defines the method for generating nodes' interests. Two choices: 'node2interests', 'random'. Default setting is False -> 'node2interests
    - weighted : bool
        boolean specifying (un)weighted. Default  unweighted
    - directed : bool
        graph is (un)directed. Default  undirected
    - output : str
        Outputs path
    - seed : int
        Seed (int) for random distribution extractions
    - dimensions : int
        Number of dimensions for node2vec. Default 128
    - walk_length : int
        ength of walk per source. Default 80
    - num_walks : int
        number of walks per source. Default 10
    - window_size : int
        context size for optimization. Default 10
    - iiter : int
        number of epochs in SGD
    - workers : int
        number of parallel workers. Default   8
    - p : float
        manually set BFS parameter
    - q : float
        manually set DFS parameter
    - norm : bool
      choose if interests have to be normalize (True) or not (False)
    - transl : bool
      choose if interests have to be translated to positive axes
    - reduce : bool
        if True reduces dimensions with NMF



    Notes
    -----
    - Arrays are translated in order to have non-negative entries
    - Dimension is reduced to the number of topics using NMF, which mantains
      positivity
    '''
    if topics >= dimensions:
        print('Topics have to be less than dimensions')
        sys.exit()

    nx_graph = read_graph(weighted=weighted, graph=graph, directed=directed)

    emb = n2i_nx_graph(
                 nx_graph=nx_graph,
                 weighted=weighted, directed=directed,
                 fast=fast,
                 seed=seed,
                 dimensions=dimensions,
                 walk_length=walk_length,
                 num_walks=num_walks,
                 window_size=window_size,
                 iiter=iiter,
                 workers=workers,
                 p=p, q=q,
                 beta=beta,
                 prior=prior,
                 alpha_value=alpha_value,
                 beta_value=beta_value,
                 verbose=verbose)

    save_emb(emb=emb, path=output, verbose=verbose)


@click.command()
@click.option('--topics', metavar='K', default=15,
                    help='Number of topics in the topic model. Default 15. K<d ',
                    type=int)
@click.option('--graph', default=None,
                    help='Input path of the graph edgelist', type=str)

@click.option('--fast', is_flag=True,
                    help="defines the method for generating nodes' interests. Two choices: 'node2interests', 'random'. Default setting is False -> 'node2interests",
                    default=False)

@click.option('--weighted', is_flag=True,
                    help='boolean specifying (un)weighted. Default  unweighted', default=False)

@click.option('--directed', is_flag=True,
                    help='graph is (un)directed. Default  undirected',
                    default=False)
@click.option('--output', default=curr_path, help='Outputs path')
@click.option('--seed', help='Seed (int) for random distribution extractions',
                    type=int, required=False)


@click.option('--dimensions', metavar='d', type=int, default=128,
                    help='Number of dimensions for node2vec. Default 128')

@click.option('--walk_length', metavar='w', type=int, default=80,
                    help='length of walk per source. Default 80')

@click.option('--num_walks', metavar='nw', type=int, default=10,
                    help='number of walks per source. Default 10')

@click.option('--window_size', metavar='ws', type=int, default=10,
                    help='context size for optimization. Default 10')

@click.option('--iiter', default=1, type=int,
                  help='number of epochs in SGD. Default 1')

@click.option('--workers', type=int, default=8,
                    help='number of parallel workers. Default   8')

@click.option('--p', type=float, default=1,
                    help='manually set BFS parameter')

@click.option('--q', type=float, default=1,
                    help='manually set DFS parameter')
@click.option('--normalize', is_flag=True,
                    help='normalize embeddings',
                    default=False)
@click.option('--translate', is_flag=True,
                    help='translate embeddings to potsitive axes',
                    default=True)
@click.option('--beta', type=int, default=0,
                    help='beta parameter for Beta-VAE loss term. Default  0')
@click.option('--prior', type=str, default='half_norm',
                    help='prior function for Beta-VAE loss term. Default  half_norm')
@click.option('--alpha_val', type=float, default=2.,
                    help='alpha value for the alpha vec of the Beta prior distribution. Default 2.')
@click.option('--beta_val', type=float, default=2.,
                    help='beta value for the beta vec of the Beta prior distribution. Default 2.')
@click.option('--verbose', is_flag=True,
                    help='n2i rpivdes all prints',
                    default=False)
def main_cli(topics,
         graph, fast,
         weighted, directed,
         output, seed,
         dimensions, walk_length,
         num_walks, window_size,
         iiter, workers,
         p, q,
         beta,
         prior,
         alpha_value,
         beta_value,
         verbose):
    '''


    Creates interests vector for each node using node2vec algorithm and
    saves interests in a file.

    3 steps:
        1. finding node2vec embeddings
        2. translation to positive axes of the embeddings
        3. reduction with NMF to _topics dimensions
    '''
    n2i_main(topics=topics,
             graph=graph, fast=fast,
             weighted=weighted, directed=directed,
             output=output, seed=seed,
             dimensions=dimensions, walk_length=walk_length,
             num_walks=num_walks, window_size=window_size,
             iiter=iiter, workers=workers,
             p=p, q=q,
             beta=beta,
             prior=prior,
             alpha_value=alpha_value,
             beta_value=beta_value,
             verbose=verbose)

if __name__ == '__main__':
    main_cli()



'''
Reference implementation of node2vec.

Author: Aditya Grover

For more details, refer to the paper:
node2vec: Scalable Feature Learning for Networks
Aditya Grover and Jure Leskovec
Knowledge Discovery and Data Mining (KDD), 2016
'''

import argparse
import numpy as np
import networkx as nx
from node2vec_git.src.node2vec_class import Graph
from gensim.models import Word2Vec

def parse_args():
    '''
    Parses the node2vec arguments.
    '''
    parser = argparse.ArgumentParser(description="Run node2vec.")

    parser.add_argument('--graph', nargs='?', default='graph/karate.edgelist',
                        help='Input graph path')

    parser.add_argument('--output', nargs='?', default='emb/karate.emb',
                        help='Embeddings path')

    parser.add_argument('--dimensions', type=int, default=128,
                        help='Number of dimensions. Default is 128.')

    parser.add_argument('--walk-length', type=int, default=80,
                        help='Length of walk per source. Default is 80.')

    parser.add_argument('--num-walks', type=int, default=10,
                        help='Number of walks per source. Default is 10.')

    parser.add_argument('--window-size', type=int, default=10,
                        help='Context size for optimization. Default is 10.')

    parser.add_argument('--iiter', default=1, type=int,
                      help='Number of epochs in SGD')

    parser.add_argument('--workers', type=int, default=8,
                        help='Number of parallel workers. Default is 8.')

    parser.add_argument('--p', type=float, default=1,
                        help='Return hyperparameter. Default is 1.')

    parser.add_argument('--q', type=float, default=1,
                        help='Inout hyperparameter. Default is 1.')

    parser.add_argument('--weighted', dest='weighted', action='store_true',
                        help='Boolean specifying (un)weighted. Default is unweighted.')
    parser.add_argument('--unweighted', dest='unweighted', action='store_false')
    parser.set_defaults(weighted=False)

    parser.add_argument('--directed', dest='directed', action='store_true',
                        help='Graph is (un)directed. Default is undirected.')
    parser.add_argument('--undirected', dest='undirected', action='store_false')
    parser.set_defaults(directed=False)

    return parser.parse_args()

def read_graph(weighted, graph, directed):
    '''
    Reads the input network in networkx.
    '''
    print(graph)
    if weighted:
        G = nx.read_edgelist(graph, nodetype=int, data=(('weight',float),), create_using=nx.DiGraph())
    else:
        G = nx.read_edgelist(graph, nodetype=int, create_using=nx.DiGraph())
        for edge in G.edges():
            G[edge[0]][edge[1]]['weight'] = 1

    if not directed:
        G = G.to_undirected()

    return G

def learn_embeddings(walks, dimensions, window_size, workers, iiter):
    '''
    Learn embeddings by optimizing the Skipgram objective using SGD.
    '''
    walks = [list(map(str, walk)) for walk in walks]
    model = Word2Vec(walks, size=dimensions, window=window_size, min_count=0,
                     sg=1, workers=workers, iter=iiter)
    #model.wv.save_word2vec_format(args.output)

    return model.wv

def node2vec_main(weighted, graph, directed, p, q, num_walks, walk_length,
                  dimensions, window_size, workers, iiter):
    '''
    Pipeline for representational learning for all nodes in a graph.
    '''
    nx_G = read_graph(weighted, graph, directed)
    G = Graph(nx_G, directed, p, q)
    G.preprocess_transition_probs()
    walks = G.simulate_walks(num_walks, walk_length)
    emb_model = learn_embeddings(walks, dimensions, window_size,
                                 workers, iiter)

    return emb_model


if __name__ == "__main__":
    args = parse_args()
    node2vec_main(args)



""" Tensorflow alternative implementation of Word2Vec;
    by default we generate non-negative embeddings.
"""

import numpy as np
import tensorflow as tf
from tqdm import tqdm

import math
import random

def generate_batch(walks, batch_size, num_skips, window):
    """ Generator that, given a list of walks (each walk being a list of indexes),
        yields the `batch, labels` pairs needed to train Word2vec.

        Args:
            batch_size: Number of pairs (target, context) in each example.
            num_skips: Number of examples for each target node.
            window: Length of the context.
    """
    assert batch_size % num_skips == 0
    assert num_skips <= 2 * window
    span = 2 * window + 1  # [ window target window ]
    for walk in walks:
        for data_index in range(0, len(walk), span):
            batch = np.ndarray(shape=(batch_size), dtype=np.int32)
            labels = np.ndarray(shape=(batch_size, 1), dtype=np.int32)
            buffer = walk[data_index:data_index + span]
            for i in range(batch_size // num_skips):
                context_words = [w for w in range(span) if w != window and w < len(buffer)]
                if num_skips > len(context_words):
                    words_to_use = random.choices(context_words, k=num_skips)
                else:
                    words_to_use = random.sample(context_words, num_skips)
                for j, context_word in enumerate(words_to_use):
                    print(len(buffer))
                    print(window)
                    batch[i * num_skips + j] = buffer[window]
                    labels[i * num_skips + j, 0] = buffer[context_word]
            yield batch, labels

EPSILON = 1E-04
def beta_kl_divergence(sample, prior_alpha, prior_beta):
    mu = tf.math.reduce_mean(sample)
    var = tf.reduce_mean(tf.squared_difference(sample, mu)) + EPSILON
    observed_alpha = ((1. - mu) / var - (1. / (mu + EPSILON))) * tf.square(mu)
    observed_beta = observed_alpha * (1. / (mu + EPSILON) - 1)
    return (
        tf.lgamma(prior_alpha + prior_beta)
        - (tf.lgamma(prior_alpha) + tf.lgamma(prior_beta))
        - (tf.lgamma(observed_alpha + observed_beta + EPSILON))
        + (tf.lgamma(observed_alpha + EPSILON) + tf.lgamma(observed_beta + EPSILON))
        + (prior_alpha - observed_alpha) * (tf.digamma(prior_alpha) - tf.digamma(prior_alpha + prior_beta))
        + (prior_beta - observed_beta) * (tf.digamma(prior_beta) - tf.digamma(prior_alpha + prior_beta))
    )

class Word2vec:
    def __init__(self,
            number_of_nodes,
            alpha=0.025,
            embedding_size=128,
            cost_of_negatives=1.,
            batch_size=100,
            num_sampled=5, # Same default as Gensim.
            beta=0,
            prior='half_norm'
    ):
        """
        Build the TensorFlow operation graph.

        Args:
            number_of_nodes: vocabulary size, aka number of nodes in the graph.
            alpha: Learning rate.
            embedding_size: Number of dimensions of the embedding.
            cost_of_negatives: Multiplier for the loss function to avoid negatives. Use 0 to allow them.
            num_sampled: Number of negative examples to sample for NCF.
        """
        batch_size = None # Will use whatever batch size will be passed at runtime.
        self.embedding_size = embedding_size

        self.embeddings = tf.Variable(
            tf.random_uniform([number_of_nodes, embedding_size], 0, 1.0),
            name='embeddings'
        )

        # nce loss
        nce_weights = tf.Variable(
          tf.truncated_normal([number_of_nodes, embedding_size],
                              stddev=1.0 / math.sqrt(embedding_size)))
        nce_biases = tf.Variable(tf.zeros([number_of_nodes]))


        self.train_inputs = tf.placeholder(tf.int32, shape=[batch_size], name='input')
        self.train_labels = tf.placeholder(tf.int32, shape=[batch_size, 1], name='labels')
        embed = tf.nn.embedding_lookup(self.embeddings, self.train_inputs)


        self.nce_loss = tf.reduce_mean(
            tf.nn.nce_loss(weights=nce_weights,
                             biases=nce_biases,
                             labels=self.train_labels,
                             inputs=embed,
                             num_sampled=num_sampled,
                             num_classes=number_of_nodes))
        # non-negative constraint
        if cost_of_negatives > 0:
            self.negatives_loss = tf.nn.softsign(
                tf.reduce_sum(- tf.minimum(0., self.embeddings)))
            self.loss = self.nce_loss + cost_of_negatives * self.negatives_loss
        else:
            self.loss = self.nce_loss


        # beta-vae constraint
        mu = tf.reduce_mean(self.embeddings, axis=0, name='mu')
        log_std = tf.log(tf.math.reduce_std(self.embeddings, axis=0, name='std'))
        #print('mu: ', mu.shape, 'log_std: ', log_std.shape)

        if prior == 'half_norm':
            # kl with half-normal distribution
            kl = 0.5 * (tf.math.square(tf.exp(log_std) - 1)) - log_std
        if prior == 'norm':
            # kl with normal distribution
            kl = 0.5 * (-log_std + tf.square(mu - 10) + tf.exp(log_std) - 1)
        if prior == 'beta':
            prior_alpha = np.array([2. for _ in range(embedding_size)], dtype=np.float32)
            prior_beta = np.array([2. for _ in range(embedding_size)], dtype=np.float32)
            kl = beta_kl_divergence(self.embeddings, prior_alpha, prior_beta)


        kl_penalty = tf.reduce_sum(kl)

        #print('loss_before ', self.loss)
        self.loss += beta * kl_penalty
        #print('loss_after ', self.loss)

        self.optimizer = tf.train.GradientDescentOptimizer(learning_rate=0.1).minimize(
            self.loss
        )


    def run(self, walks, batch_size=100, iiter=4, num_skips=None, window=5, verbose=True):
        """ Run Word2Vec on the given random walks, and returns the embeddings.

        Args:
            walks: list of random walks generated by `node2vec`, each walk being a list of indexes.
            batch_size: Number of pairs (target, context) in each example.
            iiter: number of epochs.
            num_skips: Number of examples for each target node. If `None`, it uses `2 * window`.
            window: Length of the context.
            verbose: if True, prints a `tqdm` ProgressBar.

        Returns:
            The embeddings, as a number_of_nodes x embedding_size numpy matrix.
        """
        # Preprocess args.
        if num_skips is None:
            num_skips = 2 * window
        span = 2 * window + 1

        # Initialize Training.
        with tf.Session() as session:
            session.run(tf.global_variables_initializer())

            if verbose:
                examples_per_epoch = sum(math.ceil(len(walk) / span) for walk in walks)
                pbar = tqdm(total=(iiter * examples_per_epoch), desc="Learning embeddings")

            # Training.
            for _epoch in range(iiter):
                for inputs, labels in generate_batch(walks,
                                                     batch_size=batch_size,
                                                     num_skips=num_skips,
                                                     window=window):
                    #print('inputs ', inputs)
                    feed_dict = {self.train_inputs: inputs, self.train_labels: labels}
                    _, cur_loss = session.run([
                        self.optimizer, self.loss
                    ], feed_dict=feed_dict)
                    if verbose:
                        pbar.update(1)
                        pbar.set_postfix(loss=cur_loss)

            if verbose: pbar.close()

            # Returns embeddings.
            return self.embeddings.eval()



import numpy as np
import networkx as nx
import random
from tqdm import tqdm


class Graph():
    def __init__(self, nx_G, is_directed, p, q):
        self.G = nx_G
        self.is_directed = is_directed
        self.p = p
        self.q = q

    def node2vec_walk(self, walk_length, start_node):
        '''
        Simulate a random walk starting from start node.
        '''
        G = self.G
        alias_nodes = self.alias_nodes
        alias_edges = self.alias_edges

        walk = [start_node]

        while len(list(walk)) < walk_length:
            cur = walk[-1]
            cur_nbrs = sorted(G.neighbors(cur))
            if len(list(cur_nbrs)) > 0:
                if len(list(walk)) == 1:
                    walk.append(cur_nbrs[alias_draw(alias_nodes[cur][0], alias_nodes[cur][1])])
                else:
                    prev = walk[-2]
                    next = cur_nbrs[alias_draw(alias_edges[(prev, cur)][0],
                        alias_edges[(prev, cur)][1])]
                    walk.append(next)
            else:
                break

        return walk

    def simulate_walks(self, num_walks, walk_length):
        '''
        Repeatedly simulate random walks from each node.
        '''
        G = self.G
        walks = []
        nodes = list(G.nodes())
        print('Walk iteration:')
        for walk_iter in tqdm(range(num_walks)):
            #print(str(walk_iter+1), '/', str(num_walks))
            random.shuffle(nodes)
            for node in nodes:
                walks.append(self.node2vec_walk(walk_length=walk_length, start_node=node))
        #print(walks)
        return walks

    def get_alias_edge(self, src, dst):
        '''
        Get the alias edge setup lists for a given edge.
        '''
        G = self.G
        p = self.p
        q = self.q

        unnormalized_probs = []
        for dst_nbr in sorted(G.neighbors(dst)):
            if dst_nbr == src:
                unnormalized_probs.append(G[dst][dst_nbr]['weight']/p)
            elif G.has_edge(dst_nbr, src):
                unnormalized_probs.append(G[dst][dst_nbr]['weight'])
            else:
                unnormalized_probs.append(G[dst][dst_nbr]['weight']/q)
        norm_const = sum(unnormalized_probs)
        normalized_probs =  [float(u_prob)/norm_const for u_prob in unnormalized_probs]

        return alias_setup(normalized_probs)

    def preprocess_transition_probs(self):
        '''
        Preprocessing of transition probabilities for guiding the random walks.
        '''
        G = self.G
        is_directed = self.is_directed

        alias_nodes = {}
        for node in tqdm(G.nodes()):
            unnormalized_probs = [G[node][nbr]['weight'] for nbr in sorted(G.neighbors(node))]
            norm_const = sum(unnormalized_probs)
            normalized_probs =  [float(u_prob)/norm_const for u_prob in unnormalized_probs]
            alias_nodes[node] = alias_setup(normalized_probs)

        alias_edges = {}
        triads = {}

        if is_directed:
            for edge in tqdm(G.edges()):
                alias_edges[edge] = self.get_alias_edge(edge[0], edge[1])
        else:
            for edge in tqdm(G.edges()):
                alias_edges[edge] = self.get_alias_edge(edge[0], edge[1])
                alias_edges[(edge[1], edge[0])] = self.get_alias_edge(edge[1], edge[0])

        self.alias_nodes = alias_nodes
        self.alias_edges = alias_edges

        return


def alias_setup(probs):
    '''
    Compute utility lists for non-uniform sampling from discrete distributions.
    Refer to https://hips.seas.harvard.edu/blog/2013/03/03/the-alias-method-efficient-sampling-with-many-discrete-outcomes/
    for details
    '''
    K = len(probs)
    q = np.zeros(K)
    J = np.zeros(K, dtype=np.int)

    smaller = []
    larger = []
    for kk, prob in enumerate(probs):
        q[kk] = K*prob
        if q[kk] < 1.0:
            smaller.append(kk)
        else:
            larger.append(kk)

    while len(smaller) > 0 and len(larger) > 0:
        small = smaller.pop()
        large = larger.pop()

        J[small] = large
        q[large] = q[large] + q[small] - 1.0
        if q[large] < 1.0:
            smaller.append(large)
        else:
            larger.append(large)

    return J, q

def alias_draw(J, q):
    '''
    Draw sample from a non-uniform discrete distribution using alias sampling.
    '''
    K = len(J)

    kk = int(np.floor(np.random.rand()*K))
    if np.random.rand() < q[kk]:
        return kk
    else:
        return J[kk]



'''
Main of the extended version of womg
'''
import click
from womg_core import womgc
from womg_core.__main__ import save
from womg_core.utils.distributions import set_seed
from womg.topic.lda_extended import LDAExtended



def womg_main(graph=None, #pylint: disable=too-many-arguments, too-many-locals
              docs_path=None,
              items_descr=None,
              numb_topics=15,
              numb_docs=None,
              numb_steps=6,
              homophily=.5,
              gn_strength=13.,
              infl_strength=12.,
              virality_exp=8.,
              virality_resistance=13.,
              interests=None,
              int_mode='nmf',
              weighted=False, directed=False,
              path_out=None,
              seed=None,
              progress_bar=True,
              single_activator=False):
    '''


    --------------------------------------------------------------------
    WoMG main function:

    The *WoMG* software generates synthetic datasets of documents cascades on network.
    It starts with any (un)directed, (un)weighted graph and a collection of
    documents and it outputs the propagation DAGs of the docs through the network.
    Diffusion process is guided by the nodes underlying preferences.
    Please check the github page for more details.


    Parameters
    ----------
    numb_topics : int
        number of topics in the topic model. Default 15. K<d

    numb_docs : int
        number of docs to be generated. Default 100

    numb_steps : int
        number of time steps for diffusion

    homophily : float
        0<=H<=1 :degree of homophily decoded from the given network.
        Default 0.5

    gn_strength : float
        god node influence stength. Default 13

    infl_strength : float
        relative strength of influence with respect to the interests
        Default: 12.

    virality : float
        exponent of the powerlaw distribution for documents viralities.
        P(x; a) = x^{-a}, 0 <= x <=1. Deafault a=1


    int_mode : str
        defines the method for generating nodes' interests.
        4 choices: 'n2i', 'rand', 'prop_int', 'nmf'
        Default setting is 'nmf'


    graph : str or nx obj
        input path of the graph edgelist or nx object. If nx obj, womg creates a
        copy of the input graph.

    weighted : bool
        boolean specifying (un)weighted. Default  unweighted

    directed : bool
        graph is (un)directed. Default  undirected


    docs_path : str
        input  path of the documents folder

    path_out : str
        outputs path


    seed : int
        seed (int) for random distribution extractions


    progress_bar : bool
        boolean for specifying the progress bar related to the environment
        if True progress_bar=tqdm_notebook -> Jupyter progress_bar;
        if False progress_bar=tqdm. Default False
    '''

    set_seed(seed)
    topic_model = LDAExtended(numb_topics=numb_topics,
                              numb_docs=numb_docs,
                              docs_path=docs_path,
                              items_descr=items_descr,
                              progress_bar=progress_bar)
    topic_model.fit()
    prop = womgc(graph=graph,
                 numb_topics=numb_topics,
                 numb_docs=numb_docs,
                 homophily=homophily,
                 weighted=weighted,
                 directed=directed,
                 interests=interests,
                 gn_strength=gn_strength,
                 infl_strength=infl_strength,
                 int_mode=int_mode,
                 seed=seed,
                 items_descr=topic_model.items_descript,
                 virality_exp=virality_exp,
                 single_activator=single_activator,
                 virality_resistance=virality_resistance,
                 numb_steps=numb_steps,
                 progress_bar=progress_bar,
                 path_out=path_out)
    prop.topic_model.items_keyw = topic_model.items_keyw

    return prop

@click.command()
@click.option('--topics', metavar='K', default=15,
              help='Number of topics in the topic model. Default 15. K<d ',
              type=int)
@click.option('--docs', metavar='D', default=None,
              help='Number of docs to be generated. Default 100',
              type=int)
@click.option('--steps', metavar='T', default=6,
              help='Number of time steps for diffusion',
              type=int)
@click.option('--homophily', metavar='H', default=0.5,
              help='0<=H<=1 :degree of homophily decoded from the given network. Default 0.5',
              type=click.FloatRange(0, 1, clamp=True))
@click.option('--gn_strength', default=13.,
              help='Influence strength of the god node for initial configuration. Default 13.',
              type=float)
@click.option('--infl_strength', type=float, default=None,
              help='Influence strength of nodes with respect to interests vecs. Default 12.')
@click.option('--virality_exp', metavar='V', default=8.,
              help='Exponent of the pareto distribution for documents viralities. Default 8.',
              type=float)
@click.option('--virality_resistance', metavar='V', default=13.,
              help='Virality resistance factor r. Default 13.',
              type=float)

@click.option('--graph', default=None,
              help='Input path of the graph edgelist or nx object', type=str)
@click.option('--interests', default=None,
              help='Input path of the ginterests table', type=str)

@click.option('--int_mode', type=str,
              help="defines the method for generating nodes' interests. 2 choices: 'rand', 'nmf'. Default 'nmf' ",
              default='nmf')

@click.option('--weighted', is_flag=True,
              help='boolean specifying (un)weighted. Default  unweighted', default=False)

@click.option('--directed', is_flag=True,
              help='graph is (un)directed. Default  undirected',
              default=False)


@click.option('--docs_folder', metavar='DOCS', default=None,
              help='Input  path of the documents folder', type=str)
@click.option('--items_descr', default=None,
              help='Input  path items description file representing each item in the topics space. Format: topic_index [topic-dim vec]', type=str)
@click.option('--output', default=None, help='Outputs path')
@click.option('--seed', help='Seed (int) for random distribution extractions',
              type=int, required=False)


@click.option('--progress_bar', is_flag=True,
              help='boolean for specifying the progress bar related to the environment if True progress_bar=tqdm_notebook -> Jupyter progress_bar; if False progress_bar=tqdm. Default False ',
              default=True)

@click.option('--save_int', is_flag=True,
              help='if True WoMG saves the interests vector for each node',
              default=False)
@click.option('--save_infl', is_flag=True, #pylint: disable=too-many-arguments, too-many-locals
              help='if True WoMG saves the influence vector for each node',
              default=False)
@click.option('--save_keyw', is_flag=True,
              help='if True WoMG saves the keywords in a bow format for each document',
              default=False)
@click.option('--single_activator', is_flag=True,
              help='if True we have at most one activator per item, else god node will activate all nodes beyond threshold',
              default=False)

def main_cli(graph,
             items_descr,
             topics, docs, steps, homophily,
             virality_exp,
             virality_resistance,
             interests,
             int_mode,
             weighted, directed,
             docs_folder,
             gn_strength,
             infl_strength,
             output, seed,
             progress_bar,
             single_activator,
             save_int,
             save_infl,
             save_keyw):
    '''


    The *WoMG* software generates synthetic datasets of documents cascades on network.
    It starts with any (un)directed, (un)weighted graph and a collection of
    documents and it outputs the propagation DAGs of the docs through the network.
    Diffusion process is guided by the nodes underlying preferences.
    Please check the github page for more details.

    '''
    prop = womg_main(graph=graph,
                     docs_path=docs_folder,
                     items_descr=items_descr,
                     numb_topics=topics, numb_docs=docs,
                     numb_steps=steps, homophily=homophily,
                     gn_strength=gn_strength,
                     infl_strength=infl_strength,
                     virality_exp=virality_exp,
                     virality_resistance=virality_resistance,
                     interests=interests,
                     int_mode=int_mode,
                     weighted=weighted, directed=directed,
                     path_out=output,
                     seed=seed,
                     progress_bar=progress_bar,
                     single_activator=single_activator)

    save(network_model=prop.network_model,
         topic_model=prop.topic_model,
         diffusion_model=prop.diffusion_model,
         path=output,
         save_int=save_int,
         save_infl=save_infl,
         save_keyw=save_keyw)

if __name__ == '__main__':
    main_cli() # pylint: disable=no-value-for-parameter



'''
Reference implementation of node2vec.

Author: Aditya Grover

For more details, refer to the paper:
node2vec: Scalable Feature Learning for Networks
Aditya Grover and Jure Leskovec
Knowledge Discovery and Data Mining (KDD), 2016
'''

from womg.n2i.tfoptimizer import Word2vec as TfWord2vec

import pathlib
import networkx as nx
from womg.n2i.graph import Graph
import numpy as np
from gensim.models import Word2Vec as GensimWord2Vec

def read_graph(weighted, graph, directed):
    '''
    Reads the input network in networkx.
    '''
    #print(graph)
    if weighted:
        G = nx.read_edgelist(graph, nodetype=int, data=(('weight',float),), create_using=nx.DiGraph())
    else:
        G = nx.read_edgelist(graph, nodetype=int, create_using=nx.DiGraph())
        for edge in G.edges():
            G[edge[0]][edge[1]]['weight'] = 1

    if not directed:
        G = G.to_undirected()

    return G

def learn_embeddings(number_of_nodes, walks, dimensions, window_size, workers, iiter, use_tf=True,
    beta=0, prior='half_norm',
    verbose=True):
    '''
    Learn embeddings by optimizing the Skipgram objective using SGD.
    '''
    if use_tf:
        model = TfWord2vec(number_of_nodes, embedding_size=dimensions, beta=beta, prior=prior)
        return model.run(walks, iiter=iiter, window=window_size, verbose=verbose)
    else:
        walks = [list(map(str, walk)) for walk in walks]
        model = GensimWord2Vec(walks, size=dimensions, window=window_size, min_count=0,
                         sg=1, workers=workers, iter=iiter)

        embeddings = []
        for node in range(number_of_nodes):
            curr_emb = np.array([model.wv[str(node)][_] for _ in range(dimensions)])
            embeddings.append(curr_emb)
        return np.array(embeddings)

def node2vec(weighted, graph, directed, p, q, num_walks, walk_length,
             dimensions, window_size, workers, iiter, verbose, use_tf=True,
             beta=0, prior='half_norm', alpha_value=2., beta_value=2.):
    '''
    Pipeline for representational learning for all nodes in a graph.
    '''
    if type(graph) == str:
        nx_G = read_graph(weighted, graph, directed)
    else:
        nx_G = graph

    G = Graph(nx_G, directed, p, q, verbose=verbose)
    G.preprocess_transition_probs()

    walks = G.simulate_walks(num_walks, walk_length)
    '''
    g2v = Node2Vec()
    walks = g2v.simulate_walks(nx_G, walklen=walk_length,
                                epochs=num_walks,
                                return_weight=1/p,
                                neighbor_weight=1/q,
                                threads=workers)
    #print(walks)
    '''
    emb_model = learn_embeddings(nx_G.number_of_nodes(), walks, dimensions, window_size,
                                 workers, iiter, verbose=verbose, use_tf=use_tf, beta=beta,
                                 prior=prior)

    return emb_model

def save_emb(emb, path, verbose=False):
    '''
    saves embeddings in path with the following format :
    node_id [embeddings]
    '''
    file_name = make_filename('emb', pathlib.Path(path))
    with open(file_name, 'w') as f:
        for node in emb.keys():
            tmp_emb = [_ for _ in emb[node]]
            f.write(str(node) + ' ' + str(tmp_emb) + '\n')
    if verbose:
        print('Embeddings have been saved in ', path)

def make_filename(name, output_dir, new_file=True):
    '''
    Returns the correct filename checking the file "name" is in "output_dir"
    If the file already exists it returns a filename with an incrementing
    index.

    Parameters
    ----------
    - name : str
        name of the file
    - output_dir : str
        path of the "name" file

    Returns
    -------
    Filename (str)
    '''
    pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)
    filename = output_dir / str(name + "0.txt")
    sim_numb = 0

    while pathlib.Path(filename).exists():
        sim_numb+=1
        filename = output_dir / str(name + str(sim_numb) + ".txt")
    if new_file:
        return filename
    else:
        sim_numb-=1
        filename = output_dir / str(name + str(sim_numb) + ".txt")
        return filename



import setuptools
from glob import glob

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements=['click==7.1.2', 'decorator==4.4.2', 'joblib==1.0.1',
                          'networkx==2.5', 'numpy==1.19.5', 'scipy==1.5.4',
                          'scikit-learn==0.24.1', 'tqdm==4.56.2',
                          'gensim==3.8.3', 'womg_core==1.0.4']


def find_files(folder, rec=True):
    ''' Finds all pickle, csv, txt, nx files
    '''
    list_files = list(glob(str(folder)+'/**', recursive=rec))
    files = []
    for file in list_files:
        if file.endswith('txt') or file.endswith('csv') or file.endswith('pickle') or file.endswith('nx'):
            files.append(("/"+str(file[:file.rfind('/')]), [file]) )
        else:
            continue
    return files

###############################################################################
setuptools.setup(
    name="womg",
    version="1.0.4",
    author="Federico Cinus",
    author_email="federico.cinus@isi.it",
    description="Word-of-Mouth cascades Generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FedericoCinus/WoMG.git",
    packages=setuptools.find_packages(),
    #include_package_data=True,
    #package_data = {'data': ['data/*']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    data_files   = find_files('womgdata'),
    install_requires=requirements,
    py_modules=['womg'],
    entry_points='''
        [console_scripts]
        womg=womg.__main__:main_cli
    ''',
    #scripts=['bin/womg'],
    python_requires='>=3.8',
)
